/* BeOS driver for Ghostscript using a BBitmap for buffering.
 * This driver has been heavily rewritten from the previous version.
 *
 * Copyright 1997, Jake Hamby
 * Send all comments, bug fixes, etc. to:  jehamby@lightside.com
 */


#include <View.h>
#include <Window.h>
#include <Bitmap.h>

typedef void* HWND;          /* For old DLL interface */
HWND hwndtext;               /* currently unused */

extern "C" void gsdll_draw(unsigned char *device, BView *view, BRect dest, BRect src);

extern "C" {
#define __PROTOTYPES__		/* C++ needs prototypes, of course */
#include "gx.h"			/* for gx_bitmap; includes std.h */
#include "math_.h"
#include "memory_.h"
#include "gserrors.h"
#include "gsparam.h"
#include "gxdevice.h"
#include "gdevbe.h"
#include "gsdll.h"
}

/* Procedures */

private dev_proc_open_device(be_open);
private dev_proc_sync_output(be_sync);
private dev_proc_output_page(be_output_page);
private dev_proc_close_device(be_close);
private dev_proc_map_rgb_color(be_map_rgb_color);
private dev_proc_map_color_rgb(be_map_color_rgb);
private dev_proc_fill_rectangle(be_fill_rectangle);
private dev_proc_copy_mono(be_copy_mono);
private dev_proc_copy_color(be_copy_color);
private dev_proc_draw_line(be_draw_line);
private dev_proc_strip_tile_rectangle(be_strip_tile_rectangle);
private dev_proc_get_initial_matrix(be_get_initial_matrix);
private dev_proc_get_bits(be_get_bits);
private dev_proc_get_page_device(be_get_page_device);

/* The device descriptor */
private gx_device_procs be_procs = {
	be_open,
	gx_default_get_initial_matrix,
	be_sync,
	be_output_page,
	be_close,
	be_map_rgb_color,
	be_map_color_rgb,
	be_fill_rectangle,
	NULL,			/* tile_rectangle */
	be_copy_mono,
	be_copy_color,
	be_draw_line,
	NULL,			/* get_bits */
	NULL,			/* get_params */
	NULL,			/* put_params */
	NULL,			/* map_cmyk_color */
	NULL,			/* get_xfont_procs */
	NULL,			/* get_xfont_device */
	NULL,			/* map_rgb_alpha_color */
	gx_page_device_get_page_device,
	NULL,			/* get_alpha_bits */
	NULL,			/* copy_alpha */
	NULL,			/* get_band */
	NULL,			/* copy_rop */
	NULL,			/* fill_path */
	NULL,			/* stroke_path */
	NULL,			/* fill_mask */
	NULL,			/* fill_trapezoid */
	NULL,			/* fill_parallelogram */
	NULL,			/* fill_triangle */
	NULL,			/* draw_thin_line */
	NULL,			/* begin_image */
	NULL,			/* image_data */
	NULL,			/* end_image */
	NULL			/* be_strip_tile_rectangle */
};

/* The instance is public. */
gx_device_be gs_bebox_device = {
	std_device_color_body(gx_device_be, &be_procs, "bebox",
	  INITIAL_WIDTH, INITIAL_HEIGHT,
	  INITIAL_RESOLUTION, INITIAL_RESOLUTION,
	  /*dci_color(*/24, 255, 256/*)*/),
	{ 0 },			/* std_procs */
	-1,				/* mutex */
	NULL,			/* main bitmap */
	NULL,			/* offscreen view */
	NULL,			/* bmapcache ptr */
	0,				/* next bitmap to use */
};

/* The BBitmap cache struct */
struct bmapcache_s {
	gx_bitmap_id id;
	struct bmapcache_s *next;
	BBitmap bitmap;
	
	bmapcache_s(gx_bitmap_id id, BRect bounds, color_space mode) :
		id(id), bitmap(bounds, mode), next(NULL) { }
};

/* Open the device. */
private int
be_open(gx_device *dev)
{
    gx_device_be *bdev = (gx_device_be *)dev;
	bdev->bmapcache_current = &(bdev->bmapcache);
	bdev->mutex = create_sem(1, "gsdll sem");
	
	if (dev->width == INITIAL_WIDTH)
		dev->width = (int)(8.5 * dev->x_pixels_per_inch);
	if (dev->height == INITIAL_HEIGHT)
		dev->height = (int)(11.0 * dev->y_pixels_per_inch);

	BRect frame(0.0, 0.0, dev->width-1, dev->height-1);
	bdev->m_bitmap = new BBitmap(frame, B_RGB_32_BIT, true);
	bdev->m_view = new BView(frame, "bitmap view", B_FOLLOW_NONE, 0);
	bdev->m_bitmap->AddChild(bdev->m_view);
	bdev->m_view->Window()->Lock();

#ifdef PROFILE
	bdev->fillrect = bdev->copymono = bdev->drawline = 0;
#endif

	/* notify caller about new device */
	(*pgsdll_callback)(GSDLL_DEVICE, (char *)dev, 1);
	(*pgsdll_callback)(GSDLL_SIZE, (char *)dev,
			(dev->width & 0xffff) +
			((ulong)(dev->height & 0xffff) << 16));
    return 0;
}

/* Close the device. */
private int
be_close(gx_device *dev)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    /* wait until bitmap is not being used by caller */
    gsdll_lock_device((unsigned char*)dev, 1);
	(*pgsdll_callback)(GSDLL_DEVICE, (char *)dev, 0);
	gsdll_lock_device((unsigned char*)dev, 0);

#ifdef PROFILE
    fprintf(stderr, "fillrect=%d copymono=%d drawline=%d\n",
    	bdev->fillrect, bdev->copymono, bdev->drawline);
#endif

	delete bdev->m_bitmap;	/* this also deletes m_view */
	bmapcache_s *cur2 = bdev->bmapcache;
	while(cur2) {
		bmapcache_s *next = cur2->next;
		delete cur2;
		cur2 = next;
	}

	delete_sem(bdev->mutex);
    return 0;
}

/* Map a color.  The "device colors" are rgb_color structs. */
private gx_color_index
be_map_rgb_color(register gx_device *dev,
		gx_color_value r, gx_color_value g, gx_color_value b)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    union {
		rgb_color r;
		gx_color_index g;
	} colorcast;
	colorcast.r.red = r >> 8;
	colorcast.r.green = g >> 8;
	colorcast.r.blue = b >> 8;
	colorcast.r.alpha = 0;
	return colorcast.g;
}


/* Map a "device color" back to r-g-b. */
/* Foreground and background may be mapped to other colors, so */
/* they are handled specially. */
private int
be_map_color_rgb(register gx_device *dev, gx_color_index color,
		gx_color_value prgb[3])
{
    gx_device_be *bdev = (gx_device_be *)dev;
    union {
		rgb_color r;
		gx_color_index g;
	} colorcast;
	colorcast.g = color;
	prgb[0] = colorcast.r.red << 8;
	prgb[1] = colorcast.r.green << 8;
	prgb[2] = colorcast.r.blue << 8;
	return 0;
}

/* Synchronize the display with the commands already given */
private int
be_sync(register gx_device *dev)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    bdev->m_view->Sync();
    (*pgsdll_callback)(GSDLL_SYNC, (char *)dev, 0);
    return 0;
}

/* Make the window visible, and display the output. */
int
be_output_page(gx_device *dev, int copies, int flush)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    bdev->m_view->Sync();
    (*pgsdll_callback)(GSDLL_PAGE, (char *)dev, 0);
#ifdef PROFILE
    fprintf(stderr, "fillrect=%d copymono=%d drawline=%d\n",
    	bdev->fillrect, bdev->copymono, bdev->drawline);
#endif
    return 0;
}

/* Fill a rectangle with a color. */
private int
be_fill_rectangle(register gx_device *dev,
		 int x, int y, int w, int h, gx_color_index color)
{
    gx_device_be *bdev = (gx_device_be *)dev;
#ifdef PROFILE
    bigtime_t start = system_time();
#endif
	union {
		rgb_color r;
		gx_color_index g;
	} colorcast;
	colorcast.g = color;
	bdev->m_view->SetHighColor(colorcast.r);
	bdev->m_view->FillRect(BRect(x, y, w+x-1, y+h-1));

#if 0
	if(x==0 && y==0 && w==bdev->width && h==bdev->height) {
		/* Just cleared the screen so clear the bitmap cache */
fprintf(stderr, "clear bitmap cache\n");
		bmapcache_s *cur2 = bdev->bmapcache;
		while(cur2) {
			bmapcache_s *next = cur2->next;
			delete cur2;
			cur2 = next;
		}
		bdev->bmapcache = NULL;
		bdev->bmapcache_current = &(bdev->bmapcache);
	}
#endif

#ifdef PROFILE
	bdev->fillrect += (int)(system_time() - start);
#endif
	return 0;
}

/* Copy a monochrome bitmap. */
private int
be_copy_mono(register gx_device *dev,
	    const byte *base, int sourcex, int raster, gx_bitmap_id id,
	    int x, int y, int w, int h,
	    gx_color_index zero, gx_color_index one)
{
int *ptr = (int *)base, i, j;
unsigned char *cptr = (unsigned char *)base;
    gx_device_be *bdev = (gx_device_be *)dev;
#ifdef PROFILE
    bigtime_t start = system_time();
#endif

    if(raster) {
	union {
		rgb_color r;
		gx_color_index g;
	} colorcast;
	colorcast.g = one;
	bdev->m_view->SetHighColor(colorcast.r);
	if(zero == gx_no_color_index)
		bdev->m_view->SetDrawingMode(B_OP_OVER);
	else {
		colorcast.g = zero;
		bdev->m_view->SetLowColor(colorcast.r);
		bdev->m_view->SetDrawingMode(B_OP_COPY);
	}

	bool found=false;
	BBitmap *bitmap;
	if(id != gx_no_bitmap_id) {
		bmapcache_s *current = bdev->bmapcache;
		while(current) {
		    if(current->id == id) {
		    	bitmap = &(current->bitmap);
				found=true;
				break;
		    }
			current = current->next;
		}
	}
	if(!found) {
		bmapcache_s *bmapc = new bmapcache_s(id, BRect(0.0, 0.0, (raster << 3) - 1, h-1), B_MONOCHROME_1_BIT);
	    bmapc->bitmap.SetBits(base, raster * h, 0, B_MONOCHROME_1_BIT);
		*(bdev->bmapcache_current) = bmapc;
		bdev->bmapcache_current = &(bmapc->next);
		bitmap = &(bmapc->bitmap);
	}
	bdev->m_view->DrawBitmapAsync(bitmap, BRect(sourcex, 0, sourcex+w-1, h-1), BRect(x, y, x+w-1, y+h-1));
    }
#ifdef PROFILE
	bdev->copymono += (int)(system_time() - start);
#endif
    return 0;
}

/* Copy a color bitmap. */
/* Note:  I couldn't find any PS file which uses this function, so I left
 * it unimplemented for now.  -jeh
 */
private int
be_copy_color(register gx_device *dev,
	     const byte *base, int sourcex, int raster, gx_bitmap_id id,
	     int x, int y, int w, int h)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    fprintf(stderr, "copy color!\n");
    return 0;
}

/* Draw a line */
private int
be_draw_line(register gx_device *dev,
  	     int x0, int y0, int x1, int y1, gx_color_index color)
{
    gx_device_be *bdev = (gx_device_be *)dev;
#ifdef PROFILE
    bigtime_t start = system_time();
#endif
	union {
		rgb_color r;
		gx_color_index g;
	} colorcast;
	colorcast.g = color;
	bdev->m_view->SetHighColor(colorcast.r);
	bdev->m_view->StrokeLine(BPoint(x0, y0), BPoint(x1, y1));
#ifdef PROFILE
	bdev->drawline += (int)(system_time() - start);
#endif
    return 0;
}

/* Tile a rectangle. */
/* Note:  Because the Be GUI always has at least 256 colors, and because
 *   this function is only used for halftoning, I left it unimplemented for
 *   now.  -jeh
 */
private int
be_strip_tile_rectangle(register gx_device *dev, const gx_strip_bitmap *tiles,
		       int x, int y, int w, int h,
		       gx_color_index zero, gx_color_index one,
		       int px, int py)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    fprintf(stderr, "Striptile rectangle\n");
    return 0;
}

/* Lock to make DLL threadsafe */
int gsdll_lock_device(unsigned char *dev, int flag) {
	gx_device_be *bdev = (gx_device_be *)dev;
    if (flag) {
    	acquire_sem(bdev->mutex);
    	return 1;
    }
    release_sem(bdev->mutex);
    return 0;
}

/* Draw to on-screen BView */
void gsdll_draw(unsigned char *dev, BView *view, BRect dest, BRect src) {
	gx_device_be *bdev = (gx_device_be *)dev;
	view->DrawBitmapAsync(bdev->m_bitmap, src, dest);
}